# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :data:`iris.analysis.PERCENTILE` aggregator."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import numpy.ma as ma

from iris.analysis import WPERCENTILE


class Test_aggregate(tests.IrisTest):
    def test_missing_mandatory_kwargs(self):
        emsg = (
            "weighted_percentile aggregator requires "
            ".* keyword argument 'percent'"
        )
        with self.assertRaisesRegex(ValueError, emsg):
            WPERCENTILE.aggregate("dummy", axis=0, weights=None)
        emsg = (
            "weighted_percentile aggregator requires "
            ".* keyword argument 'weights'"
        )
        with self.assertRaisesRegex(ValueError, emsg):
            WPERCENTILE.aggregate("dummy", axis=0, percent=50)

    def test_wrong_weights_shape(self):
        data = np.arange(11)
        weights = np.ones(10)
        emsg = "_weighted_percentile: weights wrong shape."
        with self.assertRaisesRegex(ValueError, emsg):
            WPERCENTILE.aggregate(data, axis=0, percent=50, weights=weights)

    def test_1d_single(self):
        data = np.arange(11)
        weights = np.ones(data.shape)
        actual = WPERCENTILE.aggregate(
            data, axis=0, percent=50, weights=weights
        )
        expected = 5
        self.assertTupleEqual(actual.shape, ())
        self.assertEqual(actual, expected)

    def test_1d_single_unequal(self):
        data = np.arange(12)
        weights = np.ones(data.shape)
        weights[0:3] = 3
        actual, weight_total = WPERCENTILE.aggregate(
            data, axis=0, percent=50, weights=weights, returned=True
        )
        expected = 2.75
        self.assertTupleEqual(actual.shape, ())
        self.assertEqual(actual, expected)
        self.assertEqual(weight_total, 18)

    def test_masked_1d_single(self):
        data = ma.arange(11)
        weights = np.ones(data.shape)
        data[3:7] = ma.masked
        actual = WPERCENTILE.aggregate(
            data, axis=0, percent=50, weights=weights
        )
        expected = 7
        self.assertTupleEqual(actual.shape, ())
        self.assertEqual(actual, expected)

    def test_1d_multi(self):
        data = np.arange(11)
        weights = np.ones(data.shape)
        percent = np.array([20, 50, 90])
        actual = WPERCENTILE.aggregate(
            data, axis=0, percent=percent, weights=weights
        )
        expected = [1.7, 5, 9.4]
        self.assertTupleEqual(actual.shape, percent.shape)
        self.assertArrayAlmostEqual(actual, expected)

    def test_1d_multi_unequal(self):
        data = np.arange(13)
        weights = np.ones(data.shape)
        weights[1::2] = 3
        percent = np.array([20, 50, 96])
        actual = WPERCENTILE.aggregate(
            data, axis=0, percent=percent, weights=weights
        )
        expected = [2.25, 6, 11.75]
        self.assertTupleEqual(actual.shape, percent.shape)
        self.assertArrayAlmostEqual(actual, expected)

    def test_masked_1d_multi(self):
        data = ma.arange(11)
        weights = np.ones(data.shape)
        data[3:9] = ma.masked
        percent = np.array([25, 50, 75])
        actual = WPERCENTILE.aggregate(
            data, axis=0, percent=percent, weights=weights
        )
        expected = [0.75, 2, 9.25]
        self.assertTupleEqual(actual.shape, percent.shape)
        self.assertArrayAlmostEqual(actual, expected)

    def test_2d_single(self):
        shape = (2, 11)
        data = np.arange(np.prod(shape)).reshape(shape)
        weights = np.ones(shape)
        actual = WPERCENTILE.aggregate(
            data, axis=0, percent=50, weights=weights
        )
        self.assertTupleEqual(actual.shape, shape[-1:])
        expected = np.arange(shape[-1]) + 5.5
        self.assertArrayEqual(actual, expected)

    def test_masked_2d_single(self):
        shape = (2, 11)
        data = ma.arange(np.prod(shape)).reshape(shape)
        data[0, ::2] = ma.masked
        data[1, 1::2] = ma.masked
        weights = np.ones(shape)
        actual = WPERCENTILE.aggregate(
            data, axis=0, percent=50, weights=weights
        )
        self.assertTupleEqual(actual.shape, shape[-1:])
        expected = np.empty(shape[-1:])
        expected[1::2] = data[0, 1::2]
        expected[::2] = data[1, ::2]
        self.assertArrayEqual(actual, expected)

    def test_2d_multi(self):
        shape = (2, 10)
        data = np.arange(np.prod(shape)).reshape(shape)
        weights = np.ones(shape)
        percent = np.array([10, 50, 70, 100])
        actual = WPERCENTILE.aggregate(
            data, axis=0, percent=percent, weights=weights
        )
        self.assertTupleEqual(actual.shape, (shape[-1], percent.size))
        expected = np.tile(np.arange(shape[-1]), percent.size).astype("f8")
        expected = expected.reshape(percent.size, shape[-1]).T
        expected[:, 1:-1] += (percent[1:-1] - 25) * 0.2
        expected[:, -1] += 10.0
        self.assertArrayAlmostEqual(actual, expected)

    def test_masked_2d_multi(self):
        shape = (3, 10)
        data = ma.arange(np.prod(shape)).reshape(shape)
        weights = np.ones(shape)
        data[1] = ma.masked
        percent = np.array([10, 50, 70, 80])
        actual = WPERCENTILE.aggregate(
            data, axis=0, percent=percent, weights=weights
        )
        self.assertTupleEqual(actual.shape, (shape[-1], percent.size))
        expected = np.tile(np.arange(shape[-1]), percent.size).astype("f8")
        expected = expected.reshape(percent.size, shape[-1]).T
        expected[:, 1:-1] += (percent[1:-1] - 25) * 0.4
        expected[:, -1] += 20.0
        self.assertArrayAlmostEqual(actual, expected)

    def test_masked_2d_multi_unequal(self):
        shape = (3, 10)
        data = ma.arange(np.prod(shape)).reshape(shape)
        weights = np.ones(shape)
        weights[0] = 3
        data[1] = ma.masked
        percent = np.array([30, 50, 75, 80])
        actual, weight_total = WPERCENTILE.aggregate(
            data, axis=0, percent=percent, weights=weights, returned=True
        )
        self.assertTupleEqual(actual.shape, (shape[-1], percent.size))
        expected = np.tile(np.arange(shape[-1]), percent.size)
        expected = expected.reshape(percent.size, shape[-1]).T
        expected[:, 1:] = 2.0 * (
            (0.875 - percent[1:] / 100.0) * data[0, np.newaxis].T
            + (percent[1:] / 100.0 - 0.375) * data[-1, np.newaxis].T
        )
        self.assertArrayAlmostEqual(actual, expected)
        self.assertTupleEqual(weight_total.shape, (shape[-1],))
        self.assertArrayEqual(weight_total, np.repeat(4, shape[-1]))


class Test_name(tests.IrisTest):
    def test(self):
        self.assertEqual(WPERCENTILE.name(), "weighted_percentile")


class Test_aggregate_shape(tests.IrisTest):
    def test_missing_mandatory_kwarg(self):
        emsg_pc = (
            "weighted_percentile aggregator requires "
            ".* keyword argument 'percent'"
        )
        emsg_wt = (
            "weighted_percentile aggregator requires "
            ".* keyword argument 'weights'"
        )
        with self.assertRaisesRegex(ValueError, emsg_pc):
            WPERCENTILE.aggregate_shape(weights=None)
        with self.assertRaisesRegex(ValueError, emsg_pc):
            kwargs = dict(weights=None)
            WPERCENTILE.aggregate_shape(**kwargs)
        with self.assertRaisesRegex(ValueError, emsg_pc):
            kwargs = dict(point=10)
            WPERCENTILE.aggregate_shape(**kwargs)
        with self.assertRaisesRegex(ValueError, emsg_wt):
            WPERCENTILE.aggregate_shape(percent=50)
        with self.assertRaisesRegex(ValueError, emsg_wt):
            kwargs = dict(percent=50)
            WPERCENTILE.aggregate_shape(**kwargs)
        with self.assertRaisesRegex(ValueError, emsg_wt):
            kwargs = dict(percent=50, weight=None)
            WPERCENTILE.aggregate_shape(**kwargs)

    def test_mandatory_kwarg_no_shape(self):
        kwargs = dict(percent=50, weights=None)
        self.assertTupleEqual(WPERCENTILE.aggregate_shape(**kwargs), ())
        kwargs = dict(percent=[50], weights=None)
        self.assertTupleEqual(WPERCENTILE.aggregate_shape(**kwargs), ())

    def test_mandatory_kwarg_shape(self):
        kwargs = dict(percent=(10, 20), weights=None)
        self.assertTupleEqual(WPERCENTILE.aggregate_shape(**kwargs), (2,))
        kwargs = dict(percent=range(13), weights=None)
        self.assertTupleEqual(WPERCENTILE.aggregate_shape(**kwargs), (13,))


class Test_cell_method(tests.IrisTest):
    def test(self):
        self.assertIsNone(WPERCENTILE.cell_method)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for
:func:`iris.fileformats.pp_load_rules._convert_vertical_coords`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.coords import DimCoord, AuxCoord
from iris.aux_factory import HybridPressureFactory, HybridHeightFactory
from iris.fileformats.pp import SplittableInt, STASH
from iris.fileformats.pp_load_rules import Reference, _convert_vertical_coords
from iris.tests.unit.fileformats import TestField


def _lbcode(value=None, ix=None, iy=None):
    if value is not None:
        result = SplittableInt(value, {"iy": slice(0, 2), "ix": slice(2, 4)})
    else:
        # N.B. if 'value' is None, both ix and iy must be set.
        result = SplittableInt(
            10000 + 100 * ix + iy, {"iy": slice(0, 2), "ix": slice(2, 4)}
        )
    return result


class TestLBVC001_Height(TestField):
    def _check_height(
        self,
        blev,
        stash,
        expect_normal=True,
        expect_fixed_height=None,
        dim=None,
    ):
        lbvc = 1
        lbcode = _lbcode(0)  # effectively unused in this case
        lblev, bhlev, bhrlev, brsvd1, brsvd2, brlev = (
            None,
            None,
            None,
            None,
            None,
            None,
        )
        coords_and_dims, factories = _convert_vertical_coords(
            lbcode=lbcode,
            lbvc=lbvc,
            blev=blev,
            lblev=lblev,
            stash=stash,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd1=brsvd1,
            brsvd2=brsvd2,
            brlev=brlev,
            dim=dim,
        )
        if expect_normal:
            expect_result = [
                (
                    DimCoord(
                        blev,
                        standard_name="height",
                        units="m",
                        attributes={"positive": "up"},
                    ),
                    dim,
                )
            ]
        elif expect_fixed_height:
            expect_result = [
                (
                    DimCoord(
                        expect_fixed_height,
                        standard_name="height",
                        units="m",
                        attributes={"positive": "up"},
                    ),
                    None,
                )
            ]
        else:
            expect_result = []
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)
        self.assertEqual(factories, [])

    def test_normal_height__present(self):
        self._check_height(blev=12.3, stash=STASH(1, 1, 1))

    def test_normal_height__present_vector(self):
        data = [12.3, 123.4, 1234.5]
        dim = 0
        for blev in [data, np.asarray(data)]:
            for dim_i in [dim, (dim,)]:
                self._check_height(blev=blev, stash=STASH(1, 1, 1), dim=dim_i)

    def test_normal_height__absent(self):
        self._check_height(blev=-1, stash=STASH(1, 1, 1), expect_normal=False)

    def test_normal_height__absent_vector(self):
        data = [-1, -1, -1]
        dim = 1
        for blev in [data, np.asarray(data)]:
            for dim_i in [dim, (dim,)]:
                self._check_height(
                    blev=blev,
                    stash=STASH(1, 1, 1),
                    expect_normal=False,
                    dim=dim_i,
                )

    def test_normal_height__absent_mixed_vector(self):
        data = [-1, 12.3, -1, 123.4]
        dim = 2
        for blev in [data, np.asarray(data)]:
            for dim_i in [dim, (dim,)]:
                self._check_height(
                    blev=blev,
                    stash=STASH(1, 1, 1),
                    expect_normal=False,
                    dim=dim_i,
                )

    def test_implied_height_1m5(self):
        self._check_height(
            blev=75.2,
            stash=STASH(1, 3, 236),
            expect_normal=False,
            expect_fixed_height=1.5,
        )

    def test_implied_height_1m5__vector(self):
        data = [1, 2, 3, 4]
        dim = 3
        for blev in [data, np.asarray(data)]:
            for dim_i in [dim, (dim,)]:
                self._check_height(
                    blev=blev,
                    stash=STASH(1, 3, 236),
                    expect_normal=False,
                    expect_fixed_height=1.5,
                    dim=dim_i,
                )

    def test_implied_height_10m(self):
        self._check_height(
            blev=75.2,
            stash=STASH(1, 3, 225),
            expect_normal=False,
            expect_fixed_height=10.0,
        )

    def test_implied_height_10m__vector(self):
        data = list(range(10))
        dim = 4
        for blev in [data, np.asarray(data)]:
            for dim_i in [dim, (dim,)]:
                self._check_height(
                    blev=blev,
                    stash=STASH(1, 3, 225),
                    expect_normal=False,
                    expect_fixed_height=10.0,
                    dim=dim_i,
                )


class TestLBVC002_Depth(TestField):
    def _check_depth(
        self,
        lbcode,
        lblev=23.0,
        blev=123.4,
        brlev=0.0,
        brsvd1=0.0,
        expect_bounds=True,
        expect_match=True,
        expect_mixed=False,
        dim=None,
    ):
        lbvc = 2
        stash = STASH(1, 1, 1)
        bhlev, bhrlev, brsvd2 = None, None, None
        coords_and_dims, factories = _convert_vertical_coords(
            lbcode=lbcode,
            lbvc=lbvc,
            blev=blev,
            lblev=lblev,
            stash=stash,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd1=brsvd1,
            brsvd2=brsvd2,
            brlev=brlev,
            dim=dim,
        )
        if expect_match:
            expect_result = [
                (
                    DimCoord(
                        lblev,
                        standard_name="model_level_number",
                        attributes={"positive": "down"},
                        units="1",
                    ),
                    dim,
                )
            ]
            if expect_bounds:
                brsvd1 = np.atleast_1d(brsvd1)
                brlev = np.atleast_1d(brlev)
                if expect_mixed:
                    lower = np.where(brsvd1 == brlev, blev, brsvd1)
                    upper = np.where(brsvd1 == brlev, blev, brlev)
                else:
                    lower, upper = brsvd1, brlev
                bounds = np.vstack((lower, upper)).T
                expect_result.append(
                    (
                        DimCoord(
                            blev,
                            standard_name="depth",
                            units="m",
                            bounds=bounds,
                            attributes={"positive": "down"},
                        ),
                        dim,
                    )
                )
            else:
                expect_result.append(
                    (
                        DimCoord(
                            blev,
                            standard_name="depth",
                            units="m",
                            attributes={"positive": "down"},
                        ),
                        dim,
                    )
                )
        else:
            expect_result = []
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)
        self.assertEqual(factories, [])

    def test_unbounded(self):
        self._check_depth(_lbcode(1), lblev=23.0, expect_bounds=False)

    def test_unbounded__vector(self):
        lblev = [1, 2, 3]
        blev = [10, 20, 30]
        brsvd1 = [5, 15, 25]
        brlev = [5, 15, 25]
        self._check_depth(
            _lbcode(1),
            lblev=lblev,
            blev=blev,
            brsvd1=brsvd1,
            brlev=brlev,
            expect_bounds=False,
            dim=1,
        )

    def test_unbounded__vector_no_depth(self):
        lblev = [1, 2, 3]
        blev = [10, 20, 30]
        brsvd1 = [5, 15, 25]
        brlev = [5, 15, 666]  # not all equal or all unequal!
        self._check_depth(
            _lbcode(1),
            lblev=lblev,
            blev=blev,
            brsvd1=brsvd1,
            brlev=brlev,
            expect_mixed=True,
            dim=0,
        )

    def test_bounded(self):
        self._check_depth(
            _lbcode(1), lblev=23.0, brlev=22.5, brsvd1=23.5, expect_bounds=True
        )

    def test_bounded__vector(self):
        lblev = [1, 2, 3]
        blev = [10, 20, 30]
        brsvd1 = [5, 15, 25]
        brlev = [15, 25, 35]
        self._check_depth(
            _lbcode(1),
            lblev=lblev,
            blev=blev,
            brsvd1=brsvd1,
            brlev=brlev,
            expect_bounds=True,
            dim=1,
        )

    def test_cross_section(self):
        self._check_depth(_lbcode(ix=1, iy=2), lblev=23.0, expect_match=False)

    def test_cross_section__vector(self):
        lblev = [1, 2, 3]
        blev = [10, 20, 30]
        brsvd1 = [5, 15, 25]
        brlev = [15, 25, 35]
        self._check_depth(
            _lbcode(ix=1, iy=2),
            lblev=lblev,
            blev=blev,
            brsvd1=brsvd1,
            brlev=brlev,
            expect_match=False,
            dim=1,
        )


class TestLBVC006_SoilLevel(TestField):
    def _check_soil_level(
        self, lbcode, lblev=12.3, expect_match=True, dim=None
    ):
        lbvc = 6
        stash = STASH(1, 1, 1)
        brsvd1, brlev = 0, 0
        if hasattr(lblev, "__iter__"):
            brsvd1 = [0] * len(lblev)
            brlev = [0] * len(lblev)
        blev, bhlev, bhrlev, brsvd2 = None, None, None, None
        coords_and_dims, factories = _convert_vertical_coords(
            lbcode=lbcode,
            lbvc=lbvc,
            blev=blev,
            lblev=lblev,
            stash=stash,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd1=brsvd1,
            brsvd2=brsvd2,
            brlev=brlev,
            dim=dim,
        )
        expect_result = []
        if expect_match:
            coord = DimCoord(
                lblev,
                long_name="soil_model_level_number",
                attributes={"positive": "down"},
                units="1",
            )
            expect_result = [(coord, dim)]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)
        self.assertEqual(factories, [])

    def test_normal(self):
        self._check_soil_level(_lbcode(0))

    def test_normal__vector(self):
        lblev = np.arange(10)
        self._check_soil_level(_lbcode(0), lblev=lblev, dim=0)

    def test_cross_section(self):
        self._check_soil_level(_lbcode(ix=1, iy=2), expect_match=False)

    def test_cross_section__vector(self):
        lblev = np.arange(10)
        self._check_soil_level(
            _lbcode(ix=1, iy=2), lblev=lblev, expect_match=False, dim=0
        )


class TestLBVC006_SoilDepth(TestField):
    def _check_soil_depth(
        self,
        lbcode,
        blev=0.05,
        brsvd1=0,
        brlev=0.1,
        expect_match=True,
        dim=None,
    ):
        lbvc = 6
        stash = STASH(1, 1, 1)
        lblev, bhlev, bhrlev, brsvd2 = None, None, None, None
        coords_and_dims, factories = _convert_vertical_coords(
            lbcode=lbcode,
            lbvc=lbvc,
            blev=blev,
            lblev=lblev,
            stash=stash,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd1=brsvd1,
            brsvd2=brsvd2,
            brlev=brlev,
            dim=dim,
        )
        expect_result = []
        if expect_match:
            coord = DimCoord(
                blev,
                standard_name="depth",
                bounds=np.vstack((brsvd1, brlev)).T,
                units="m",
                attributes={"positive": "down"},
            )
            expect_result = [(coord, dim)]
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)
        self.assertEqual(factories, [])

    def test_normal(self):
        self._check_soil_depth(_lbcode(0))

    def test_normal__vector(self):
        points = np.arange(10)
        self._check_soil_depth(
            _lbcode(0), blev=points, brsvd1=points - 1, brlev=points + 1, dim=0
        )

    def test_bad_bounds(self):
        points = [-0.5, 0.5]
        lower = [-1, 1]
        upper = [-1, 1]
        self._check_soil_depth(
            _lbcode(0),
            blev=points,
            brsvd1=lower,
            brlev=upper,
            dim=0,
            expect_match=False,
        )

    def test_cross_section(self):
        self._check_soil_depth(_lbcode(ix=1, iy=2), expect_match=False)

    def test_cross_section__vector(self):
        points = np.arange(10)
        self._check_soil_depth(
            _lbcode(ix=1, iy=2),
            blev=points,
            brsvd1=points - 1,
            brlev=points + 1,
            expect_match=False,
            dim=0,
        )


class TestLBVC008_Pressure(TestField):
    def _check_pressure(self, lbcode, blev=250.3, expect_match=True, dim=None):
        lbvc = 8
        stash = STASH(1, 1, 1)
        lblev, bhlev, bhrlev, brsvd1, brsvd2, brlev = (
            None,
            None,
            None,
            None,
            None,
            None,
        )
        coords_and_dims, factories = _convert_vertical_coords(
            lbcode=lbcode,
            lbvc=lbvc,
            blev=blev,
            lblev=lblev,
            stash=stash,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd1=brsvd1,
            brsvd2=brsvd2,
            brlev=brlev,
            dim=dim,
        )
        if expect_match:
            expect_result = [
                (DimCoord(blev, long_name="pressure", units="hPa"), dim)
            ]
        else:
            expect_result = []
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)
        self.assertEqual(factories, [])

    def test_normal(self):
        self._check_pressure(_lbcode(0))

    def test_normal__vector(self):
        blev = [10, 100, 1000, 10000]
        self._check_pressure(_lbcode(0), blev=blev, dim=2)

    def test_non_pressure_cross_section(self):
        self._check_pressure(_lbcode(ix=10, iy=11))

    def test_non_pressure_cross_section__vector(self):
        blev = np.arange(10)
        self._check_pressure(_lbcode(ix=10, iy=11), blev=blev, dim=0)

    def test_pressure_cross_section(self):
        self._check_pressure(_lbcode(ix=10, iy=1), expect_match=False)

    def test_pressure_cross_section__vector(self):
        blev = np.arange(10)
        self._check_pressure(
            _lbcode(ix=10, iy=1), blev=blev, dim=1, expect_match=False
        )


class TestLBVC019_PotentialTemperature(TestField):
    def _check_potm(self, lbcode, blev=130.6, expect_match=True, dim=None):
        lbvc = 19
        stash = STASH(1, 1, 1)
        lblev, bhlev, bhrlev, brsvd1, brsvd2, brlev = (
            None,
            None,
            None,
            None,
            None,
            None,
        )
        coords_and_dims, factories = _convert_vertical_coords(
            lbcode=lbcode,
            lbvc=lbvc,
            blev=blev,
            lblev=lblev,
            stash=stash,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd1=brsvd1,
            brsvd2=brsvd2,
            brlev=brlev,
            dim=dim,
        )
        if expect_match:
            expect_result = [
                (
                    DimCoord(
                        blev,
                        standard_name="air_potential_temperature",
                        units="K",
                        attributes={"positive": "up"},
                    ),
                    dim,
                )
            ]
        else:
            expect_result = []
        self.assertCoordsAndDimsListsMatch(coords_and_dims, expect_result)
        self.assertEqual(factories, [])

    def test_normal(self):
        self._check_potm(_lbcode(0))

    def test_normal__vector(self):
        blev = list(range(10))
        self._check_potm(_lbcode(0), blev=blev, dim=0)

    def test_cross_section(self):
        self._check_potm(_lbcode(ix=10, iy=11), expect_match=False)

    def test_cross_section__vector(self):
        blev = np.arange(5) + 100
        self._check_potm(
            _lbcode(ix=10, iy=11), blev=blev, dim=1, expect_match=False
        )


class TestLBVC009_HybridPressure(TestField):
    def _check(
        self,
        lblev=37.0,
        bhlev=850.1,
        bhrlev=810.0,
        brsvd2=875.0,
        blev=0.15,
        brlev=0.11,
        brsvd1=0.19,
        expect_match=True,
        dim=None,
    ):
        lbvc = 9
        lbcode = _lbcode(0)  # unused
        stash = STASH(1, 1, 1)  # unused
        coords_and_dims, factories = _convert_vertical_coords(
            lbcode=lbcode,
            lbvc=lbvc,
            blev=blev,
            lblev=lblev,
            stash=stash,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd1=brsvd1,
            brsvd2=brsvd2,
            brlev=brlev,
            dim=dim,
        )
        expect_coords_and_dims = [
            (
                DimCoord(
                    lblev,
                    standard_name="model_level_number",
                    attributes={"positive": "up"},
                    units="1",
                ),
                dim,
            )
        ]

        bhrlev = np.atleast_1d(bhrlev)
        brsvd2 = np.atleast_1d(brsvd2)
        expect_coords_and_dims.append(
            (
                DimCoord(
                    bhlev,
                    long_name="level_pressure",
                    units="Pa",
                    bounds=np.vstack((bhrlev, brsvd2)).T,
                ),
                dim,
            )
        )
        brlev = np.atleast_1d(brlev)
        brsvd1 = np.atleast_1d(brsvd1)
        expect_coords_and_dims.append(
            (
                AuxCoord(
                    blev,
                    long_name="sigma",
                    bounds=np.vstack((brlev, brsvd1)).T,
                    units="1",
                ),
                dim,
            )
        )
        expect_factories = [
            (
                HybridPressureFactory,
                [
                    {"long_name": "level_pressure"},
                    {"long_name": "sigma"},
                    Reference("surface_air_pressure"),
                ],
            )
        ]
        self.assertCoordsAndDimsListsMatch(
            coords_and_dims, expect_coords_and_dims
        )
        self.assertEqual(factories, expect_factories)

    def test_normal(self):
        self._check()

    def test_normal__vector(self):
        lblev = list(range(3))
        bhlev = [10, 20, 30]
        bhrlev = [5, 15, 25]
        brsvd2 = [15, 25, 35]
        blev = [100, 200, 300]
        brlev = [50, 150, 250]
        brsvd1 = [150, 250, 350]
        self._check(
            lblev=lblev,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd2=brsvd2,
            blev=blev,
            brlev=brlev,
            brsvd1=brsvd1,
            dim=0,
        )


class TestLBVC065_HybridHeight(TestField):
    def _check(
        self,
        lblev=37.0,
        blev=9596.3,
        brlev=9500.0,
        brsvd1=9800.0,
        bhlev=0.35,
        bhrlev=0.31,
        brsvd2=0.39,
        dim=None,
    ):
        lbvc = 65
        lbcode = _lbcode(0)  # unused
        stash = STASH(1, 1, 1)  # unused
        coords_and_dims, factories = _convert_vertical_coords(
            lbcode=lbcode,
            lbvc=lbvc,
            blev=blev,
            lblev=lblev,
            stash=stash,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd1=brsvd1,
            brsvd2=brsvd2,
            brlev=brlev,
            dim=dim,
        )
        expect_coords_and_dims = [
            (
                DimCoord(
                    lblev,
                    standard_name="model_level_number",
                    attributes={"positive": "up"},
                    units="1",
                ),
                dim,
            )
        ]
        brlev = np.atleast_1d(brlev)
        brsvd1 = np.atleast_1d(brsvd1)
        expect_coords_and_dims.append(
            (
                DimCoord(
                    blev,
                    long_name="level_height",
                    units="m",
                    bounds=np.vstack((brlev, brsvd1)).T,
                    attributes={"positive": "up"},
                ),
                dim,
            )
        )
        bhrlev = np.atleast_1d(bhrlev)
        brsvd2 = np.atleast_1d(brsvd2)
        expect_coords_and_dims.append(
            (
                AuxCoord(
                    bhlev,
                    long_name="sigma",
                    bounds=np.vstack((bhrlev, brsvd2)).T,
                    units="1",
                ),
                dim,
            )
        )
        expect_factories = [
            (
                HybridHeightFactory,
                [
                    {"long_name": "level_height"},
                    {"long_name": "sigma"},
                    Reference("orography"),
                ],
            )
        ]
        self.assertCoordsAndDimsListsMatch(
            coords_and_dims, expect_coords_and_dims
        )
        self.assertEqual(factories, expect_factories)

    def test_normal(self):
        self._check()

    def test_normal__vector(self):
        npts = 5
        lblev = np.arange(npts)
        blev = np.arange(npts) + 10
        brlev = np.arange(npts) + 5
        brsvd1 = np.arange(npts) + 15
        bhlev = np.arange(npts) + 12
        bhrlev = np.arange(npts) + 6
        brsvd2 = np.arange(npts) + 18
        self._check(
            lblev=lblev,
            blev=blev,
            brlev=brlev,
            brsvd1=brsvd1,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd2=brsvd2,
            dim=1,
        )


class TestLBVCxxx_Unhandled(TestField):
    def test_unknown_lbvc(self):
        lbvc = 999
        blev, lblev, bhlev, bhrlev, brsvd1, brsvd2, brlev = (
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )
        lbcode = _lbcode(0)  # unused
        stash = STASH(1, 1, 1)  # unused
        coords_and_dims, factories = _convert_vertical_coords(
            lbcode=lbcode,
            lbvc=lbvc,
            blev=blev,
            lblev=lblev,
            stash=stash,
            bhlev=bhlev,
            bhrlev=bhrlev,
            brsvd1=brsvd1,
            brsvd2=brsvd2,
            brlev=brlev,
        )
        self.assertEqual(coords_and_dims, [])
        self.assertEqual(factories, [])


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.plot.scatter` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests
from iris.tests.unit.plot import TestGraphicStringCoord

if tests.MPL_AVAILABLE:
    import iris.plot as iplt


@tests.skip_plot
class TestStringCoordPlot(TestGraphicStringCoord):
    def setUp(self):
        super().setUp()
        self.cube = self.cube[0, :]
        self.lat_lon_cube = self.lat_lon_cube[0, :]

    def test_xaxis_labels(self):
        iplt.scatter(self.cube.coord("str_coord"), self.cube)
        self.assertBoundsTickLabels("xaxis")

    def test_yaxis_labels(self):
        iplt.scatter(self.cube, self.cube.coord("str_coord"))
        self.assertBoundsTickLabels("yaxis")

    def test_xaxis_labels_with_axes(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlim(0, 3)
        iplt.scatter(self.cube.coord("str_coord"), self.cube, axes=ax)
        plt.close(fig)
        self.assertPointsTickLabels("xaxis", ax)

    def test_yaxis_labels_with_axes(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_ylim(0, 3)
        iplt.scatter(self.cube, self.cube.coord("str_coord"), axes=ax)
        plt.close(fig)
        self.assertPointsTickLabels("yaxis", ax)

    def test_scatter_longitude(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        iplt.scatter(
            self.lat_lon_cube, self.lat_lon_cube.coord("longitude"), axes=ax
        )
        plt.close(fig)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.coord_categorisation.add_categorised_coord`."""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

from cf_units import CALENDARS as calendars
from cf_units import Unit
import numpy as np

from iris.coord_categorisation import add_categorised_coord
from iris.coord_categorisation import add_day_of_year
from iris.cube import Cube
from iris.coords import DimCoord


class Test_add_categorised_coord(tests.IrisTest):
    def setUp(self):
        # Factor out common variables and objects.
        self.cube = mock.Mock(name="cube", coords=mock.Mock(return_value=[]))
        self.coord = mock.Mock(
            name="coord", points=np.arange(12).reshape(3, 4)
        )
        self.units = "units"
        self.vectorised = mock.Mock(name="vectorized_result")

    def test_vectorise_call(self):
        # Check that the function being passed through gets called with
        # numpy.vectorize, before being applied to the points array.
        # The reason we use numpy.vectorize is to support multi-dimensional
        # coordinate points.
        def fn(coord, v):
            return v ** 2

        with mock.patch(
            "numpy.vectorize", return_value=self.vectorised
        ) as vectorise_patch:
            with mock.patch("iris.coords.AuxCoord") as aux_coord_constructor:
                add_categorised_coord(
                    self.cube, "foobar", self.coord, fn, units=self.units
                )

        # Check the constructor of AuxCoord gets called with the
        # appropriate arguments.
        # Start with the vectorised function.
        vectorise_patch.assert_called_once_with(fn)
        # Check the vectorize wrapper gets called with the appropriate args.
        self.vectorised.assert_called_once_with(self.coord, self.coord.points)
        # Check the AuxCoord constructor itself.
        aux_coord_constructor.assert_called_once_with(
            self.vectorised(self.coord, self.coord.points),
            units=self.units,
            attributes=self.coord.attributes.copy(),
        )
        # And check adding the aux coord to the cube mock.
        self.cube.add_aux_coord.assert_called_once_with(
            aux_coord_constructor(), self.cube.coord_dims(self.coord)
        )

    def test_string_vectorised(self):
        # Check that special case handling of a vectorized string returning
        # function is taking place.
        def fn(coord, v):
            return "0123456789"[:v]

        with mock.patch(
            "numpy.vectorize", return_value=self.vectorised
        ) as vectorise_patch:
            with mock.patch("iris.coords.AuxCoord") as aux_coord_constructor:
                add_categorised_coord(
                    self.cube, "foobar", self.coord, fn, units=self.units
                )

        self.assertEqual(
            aux_coord_constructor.call_args[0][0],
            vectorise_patch(fn, otypes=[object])(
                self.coord, self.coord.points
            ).astype("|S64"),
        )


class Test_add_day_of_year(tests.IrisTest):
    def setUp(self):
        self.expected = {
            "standard": np.array(list(range(360, 367)) + list(range(1, 4))),
            "gregorian": np.array(list(range(360, 367)) + list(range(1, 4))),
            "proleptic_gregorian": np.array(
                list(range(360, 367)) + list(range(1, 4))
            ),
            "noleap": np.array(list(range(359, 366)) + list(range(1, 4))),
            "julian": np.array(list(range(360, 367)) + list(range(1, 4))),
            "all_leap": np.array(list(range(360, 367)) + list(range(1, 4))),
            "365_day": np.array(list(range(359, 366)) + list(range(1, 4))),
            "366_day": np.array(list(range(360, 367)) + list(range(1, 4))),
            "360_day": np.array(list(range(355, 361)) + list(range(1, 5))),
        }

    def make_cube(self, calendar):
        n_times = 10
        cube = Cube(np.arange(n_times))
        time_coord = DimCoord(
            np.arange(n_times),
            standard_name="time",
            units=Unit("days since 1980-12-25", calendar=calendar),
        )
        cube.add_dim_coord(time_coord, 0)
        return cube

    def test_calendars(self):
        for calendar in calendars:
            # Skip the Julian calendar due to
            # https://github.com/Unidata/netcdftime/issues/13
            # Remove this if block once the issue is resolved.
            if calendar == "julian":
                continue
            cube = self.make_cube(calendar)
            add_day_of_year(cube, "time")
            points = cube.coord("day_of_year").points
            expected_points = self.expected[calendar]
            msg = "Test failed for the following calendar: {}."
            self.assertArrayEqual(
                points, expected_points, err_msg=msg.format(calendar)
            )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Provides NAME file format loading capabilities."""

import iris.io


def _get_NAME_loader(filename):
    """
    Return the approriate load function for a NAME file based
    on the contents of its header.

    """
    # Lazy import to avoid importing name_loaders until
    # attempting to load a NAME file.
    import iris.fileformats.name_loaders as name_loaders

    load = None
    with open(filename, "r") as file_handle:
        header = name_loaders.read_header(file_handle)

    # Infer file type based on contents of header.
    if "Run name" in header and "Output format" not in header:
        if "X grid origin" not in header:
            load = name_loaders.load_NAMEIII_trajectory
        elif header.get("X grid origin") is not None:
            load = name_loaders.load_NAMEIII_field
        else:
            load = name_loaders.load_NAMEIII_timeseries

    elif "Output format" in header:
        load = name_loaders.load_NAMEIII_version2

    elif "Title" in header:
        if "Number of series" in header:
            load = name_loaders.load_NAMEII_timeseries
        else:
            load = name_loaders.load_NAMEII_field

    if load is None:
        raise ValueError(
            "Unable to determine NAME file type " "of {!r}.".format(filename)
        )

    return load


def load_cubes(filenames, callback):
    """
    Return a generator of cubes given one or more filenames and an
    optional callback.

    Args:

    * filenames (string/list):
        One or more NAME filenames to load.

    Kwargs:

    * callback (callable function):
        A function which can be passed on to :func:`iris.io.run_callback`.

    Returns:
         A generator of :class:`iris.cubes.Cube` instances.

    """
    if isinstance(filenames, str):
        filenames = [filenames]

    for filename in filenames:
        load = _get_NAME_loader(filename)
        for cube in load(filename):
            if callback is not None:
                cube = iris.io.run_callback(callback, cube, None, filename)
            if cube is not None:
                yield cube



"""
Example of a Polar Stereographic Plot
=====================================

Demonstrates plotting data that are defined on a polar stereographic
projection.

"""

import matplotlib.pyplot as plt

import iris
import iris.plot as iplt
import iris.quickplot as qplt


def main():
    file_path = iris.sample_data_path("toa_brightness_stereographic.nc")
    cube = iris.load_cube(file_path)
    qplt.contourf(cube)
    ax = plt.gca()
    ax.coastlines()
    ax.gridlines()
    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

"""Unit tests for :func:`iris.analysis.cartography._xy_range`"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.

import iris.tests as tests
import iris.tests.stock as stock
import numpy as np
from iris.analysis.cartography import _xy_range


class Test(tests.IrisTest):
    def test_bounds_mismatch(self):
        cube = stock.realistic_3d()
        cube.coord("grid_longitude").guess_bounds()

        with self.assertRaisesRegex(ValueError, "bounds"):
            _ = _xy_range(cube)

    def test_non_circular(self):
        cube = stock.realistic_3d()
        assert not cube.coord("grid_longitude").circular

        result_non_circ = _xy_range(cube)
        self.assertEqual(result_non_circ, ((-5.0, 5.0), (-4.0, 4.0)))

    @tests.skip_data
    def test_geog_cs_circular(self):
        cube = stock.global_pp()
        assert cube.coord("longitude").circular

        result = _xy_range(cube)
        np.testing.assert_array_almost_equal(
            result, ((0, 360), (-90, 90)), decimal=0
        )

    @tests.skip_data
    def test_geog_cs_regional(self):
        cube = stock.global_pp()
        cube = cube[10:20, 20:30]
        assert not cube.coord("longitude").circular

        result = _xy_range(cube)
        np.testing.assert_array_almost_equal(
            result, ((75, 108.75), (42.5, 65)), decimal=0
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""A module to provide an optimal array structure calculation."""


from iris.fileformats._structured_array_identification import GroupStructure


def _optimal_dimensioning_structure(structure, element_priorities):
    """
    Uses the structure options provided by the
    :class:`~iris.fileformats._structured_array_identification.GroupStructure`
    to determine the optimal array structure for the :class:`FieldCollation`.

    The optimal structure is that which generates the greatest number of
    non-trivial dimensions. If the number of non-trivial dimensions is equal
    in more than one structure options then dimension priorities as specified
    by `element_priorities` are used to determine optimal structure.

    Args:

    * structure:
        A set of structure options, as provided by :class:\
        `~iris.fileformats._structured_array_identification.GroupStructure`.

    * element_priorities:
        A dictionary mapping structure element names to their priority as
        defined by their input order to :func:`~optimal_array_structure`.

    Returns:

        The determined optimal array structure or an empty list if no structure
        options were determined.

    """
    permitted_structures = structure.possible_structures()
    if not permitted_structures:
        result = []
    else:
        result = max(
            permitted_structures,
            key=lambda candidate: (
                len(candidate),
                [element_priorities[name] for (name, struct) in candidate],
            ),
        )
    return result


def optimal_array_structure(ordering_elements, actual_values_elements=None):
    """
    Calculate an optimal array replication structure for a set of vectors.

    Args:

    * ordering_elements (iterable of (name, 1-d array)):
        Input element names and value-vectors.  Must all be the same length
        (but not necessarily type).  Must have at least one.

    Kwargs:

    * actual_values_elements (iterable of (name, 1-d array)):
        The 'real' values used to construct the result arrays, if different
        from 'ordering_elements'.  Must contain all the same names (but not
        necessarily in the same order).

    The 'ordering_elements' arg contains the pattern used to deduce a
    structure.  The order of this is significant, in that earlier elements get
    priority when associating dimensions with specific elements.

    Returns:

        dims_shape, primary_elements, element_arrays_and_dims, where:

        * 'dims_shape' is the shape of the vector dimensions chosen.

        * 'primary_elements' is a set of dimension names; the names of input
            elements that are identified as dimensions. At most one for each
            dimension.

        * 'element_arrays_and_dims' is a dictionary [name: (array, dims)],
            for all elements that are not dimensionless. Each array is reduced
            to the shape of its mapped dimension.

    For example::

        >>> import iris.fileformats.um._optimal_array_structuring as optdims
        >>> elements_structure = [('a', np.array([1, 1, 1, 2, 2, 2])),
        ...                       ('b', np.array([0, 1, 2, 0, 1, 2])),
        ...                       ('c', np.array([11, 12, 13, 14, 15, 16]))]
        >>> elements_values = [('a', np.array([10, 10, 10, 12, 12, 12])),
        ...                    ('b', np.array([15, 16, 17, 15, 16, 17])),
        ...                    ('c', np.array([9, 3, 5, 2, 7, 1]))]
        >>> dims_shape, dim_names, arrays_and_dims = \
        ...      optdims.optimal_array_structure(elements_structure,
        ...                                      elements_values)
        >>> print dims_shape
        (2, 3)
        >>> print dim_names
        set(['a', 'b'])
        >>> print arrays_and_dims
        {'a': (array([10, 12]), (0,)), 'c': (array([[9, 3, 5],
               [2, 7, 1]]), (0, 1)), 'b': (array([15, 16, 17]), (1,))}

    """
    # Convert the inputs to dicts.
    element_ordering_arrays = dict(ordering_elements)
    if actual_values_elements is None:
        actual_values_elements = element_ordering_arrays
    actual_value_arrays = dict(actual_values_elements)
    if set(actual_value_arrays.keys()) != set(element_ordering_arrays.keys()):
        msg = "Names in values arrays do not match those in ordering arrays."
        raise ValueError(msg)

    # Define element priorities from ordering, to choose between equally good
    # structures, as structure code does not recognise any element ordering.
    n_elements = len(ordering_elements)
    element_priorities = {
        name: n_elements - index
        for index, (name, array) in enumerate(ordering_elements)
    }

    # Calculate the basic fields-group array structure.
    base_structure = GroupStructure.from_component_arrays(
        element_ordering_arrays
    )

    # Work out the target cube structure.
    target_structure = _optimal_dimensioning_structure(
        base_structure, element_priorities
    )

    # Work out result cube dimensions.
    if not target_structure:
        # Get the length of an input array (they are all the same).
        # Note that no elements map to multiple dimensions.
        elements_length = len(ordering_elements[0][1])
        vector_dims_shape = (elements_length,)
    else:
        vector_dims_shape = tuple(
            struct.size for (_, struct) in target_structure
        )

    # Build arrays of element values mapped onto the vectorised dimensions.
    elements_and_dimensions = base_structure.build_arrays(
        vector_dims_shape, actual_value_arrays
    )

    # Filter out the trivial (scalar) ones.
    elements_and_dimensions = {
        name: (array, dims)
        for name, (array, dims) in elements_and_dimensions.items()
        if len(dims)
    }

    # Make a list of 'primary' elements; i.e. those in the target structure.
    primary_dimension_elements = set(name for (name, _) in target_structure)

    if vector_dims_shape == (1,):
        shape = ()
    else:
        shape = vector_dims_shape

    # Return all the information.
    return (shape, primary_dimension_elements, elements_and_dimensions)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.config.NetCDF` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import warnings

import iris.config


class Test(tests.IrisTest):
    def setUp(self):
        self.options = iris.config.NetCDF()

    def test_basic(self):
        self.assertFalse(self.options.conventions_override)

    def test_enabled(self):
        self.options.conventions_override = True
        self.assertTrue(self.options.conventions_override)

    def test_bad_value(self):
        # A bad value should be ignored and replaced with the default value.
        bad_value = "wibble"
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.options.conventions_override = bad_value
        self.assertFalse(self.options.conventions_override)
        exp_wmsg = "Attempting to set invalid value {!r}".format(bad_value)
        self.assertRegex(str(w[0].message), exp_wmsg)

    def test__contextmgr(self):
        with self.options.context(conventions_override=True):
            self.assertTrue(self.options.conventions_override)
        self.assertFalse(self.options.conventions_override)


if __name__ == "__main__":
    tests.main()
